<?php
/**
* Класс для отсылки сообщений
*
* Поддерживаются как сообщения в HTML, так и обычный текст.
* Письма должны быть составлены в кодировке UTF-8.
* После создания экземпляра класса (переданная в параметрах
* строка, отличная от HTML, приведёт к тому, что письмо
* будет отослано как текст), необходимо заполнить следующие
* значения:
*  [метод] <class>->AddTo($email, $txt) — для кого письмо
*  [метод] <class>->SetFrom($email, $txt) — от кого письмо
*  [поле]  <class>->subject — тема сообщения
*  [поле]  <class->body — тело сообщения
* Для отсылки письма используется метод Send().
*
* @author       andi <ab@cdef.ru>
* @copyright    © Centurys ent., 2007
* @link         http://a-panov.ru/
* @icq          288-703-918
* @since        1.0.0b
* @package      public scripts
* @access       public
*/

require_once 'canMimes.php';

/**
* canMail класс
*
* Класс служит для формирования и отсылки писем. Поддерживает вложения.
* Позволяет рассылать письма в форматах TEXT, HTML. Поддерживает вставку
* картинок (только для формата HTML).
*/
class canMail {
public static
    // допустимые расширения файлов рисунков
    $EXT = array('jpg', 'png', 'gif');
    
public
    $MAX_STRING_LENGTH  = 68,       // максимально допустимая длина строки (не включая \r\n)
    $CHUNK_LENGTH       = 45;       // длина строки (в байтах!), на которую разбивается строки при Mime-кодировании
    
private
    $_type              = '',       // тип письма: TEXT или HTML
    $_related           = '',       // "разделитель" вложений, сообщений и пр.
    
    $_build_to          = true,     // флаг; указывает, что поле «Кому» обработано
    $_build_subject     = true,     // флаг; указывает, что тема письма обработана
    $_build_body        = true,     // флаг; указывает, что тело письма обработано
    $_build_images      = true,     // флаг; указывает, что картинки обработаны
    $_build_attaches    = true,     // флаг; указывает, что вложения обработаны
    $_build_headers     = true,     // флаг; указывает, что заголовки обработаны
    
    $_a_to              = array(),  // список адресов, на которые отсылается письмо
    $_a_cc              = array(),  // список адресов, на которые отсылается копия письма
    $_a_bcc             = array(),  // список адресов, на которые отсылается скрытая копия письма
    $_a_headers         = array(),  // массив с дополнительными заголовками
    
    $_a_images          = array(),  // массив картинок
    $_a_attaches        = array(),  // массив прицепов
    
    $_reply_to          = array(),  // адрес, на который должен отправляться ответ
    $_return_path       = '',       // адрес, на который должен отправляться отбой (в случае неудачи доставки письма)
    
    $_from              = array(),  // от кого письмо
    $_subject           = '',       // тема письма
    $_body              = '',       // тело письма
    
    // отдельно формируются: To, Subject, body (body + message + attaches)
    // headers (From, Reply-To, CC, etc.)
    
    $__to               = '',       // кому (адрес) («упаковано» в base64)
    $__subject          = '',       // тема («упакованная» в base64)
    $__headers          = '',       // заголовки (сс + bcc + from + headers + reply-to + return-path)
    $__body             = '',       // тело сообщения
    $__message          = '',       // готовое сообщение (body + images + attaches)
    
    // статистика
    $_sended            = 0,        // количество отосланных писем
    $_nosended          = 0;        // количество писем, которые не удалось отослать
    
    /**
    * Создаёт объект класса canMail
    *
    * @param string $type в каком виде сформировано тело письма: в текстовом или в виде html
    */
    function __construct($type = 'TEXT') {
        $this->_type = $type === 'HTML' ? 'HTML' : 'TEXT';
        // формируем уникальный разделитель
        $this->_related = '--b' . sha1(uniqid(rand()));
    }
    
    /**
    * Добавляет изображение в текст письма
    *
    * Изображение будет добавлено только если тип письма указан как HTML.
    * Для текстовых данных это не имеет смысла, поэтому будет вызван метод
    * AddAttach(). Тип изображения проверяется по расширению файла!
    *
    * @param string $path путь к изображению
    * @param string $cid уникальный идентификатор изображения (если передано
    *   пустое значение, будет использовано имя файла)
    * @return boolean true, если изображение принято
    */
    function AddImage($path, $cid = '') {
        // для того, чтобы вложение было нормлаьно доступно, не нужно
        // использовать следующие символы: с кодом, менее 33 и /*?\:"<>|
        if (preg_match('![\x00-\x20/\*\?\\\\:"<>\|]!', $cid)) return false;
        
        // если не HTML, всё-равно прикрепим файл, только уже как вложение
        if ($this->_type !== 'HTML') return $this->AddAttach($path, $cid);
        
        if ( ! file_exists($path)) return false;
        
        $pia = pathinfo($path);
        if (!isset($pia['extension'])) $pia['extension'] = '';
        $ext = strtolower($pia['extension']);
        if (empty($ext) || !in_array($ext, self::$EXT)) return false;
        if (empty($cid)) $cid = $pia['basename'];
        
        $subtypes = canMimes::Get($ext);
        if (empty($subtypes)) return false;
        
        $this->_a_images[$cid] = array(
            'path'    => $path,
            'name'    => preg_replace('/./e', '"%".dechex(ord("$0"))', $cid),
            'subtype' => $subtypes
        );
        
        // если картинка добавлена, требуется её обработка (формирование
        // значения с ключём data, содержащим содержимое файла в виде, годном
        // для пересылки)
        $this->_build_images = false;
        return true;
    }
    
    /**
    * Добавляет вложение к письму
    */
    function AddAttach($path, $name = '') {
        if (!file_exists($path)) return false;
        
        $pia = pathinfo($path);
        if (!isset($pia['extension'])) $pia['extension'] = '';
        $ext = strtolower($pia['extension']);
        $subtypes = canMimes::Get($ext);
        if (empty($subtypes)) $subtypes = canMimes::Get();
        if (empty($name)) $name = $pia['basename'];
        if (preg_match('![\x00-\x20/\*\?\\\\:"<>\|]!', $name)) return false;
        
        $this->_a_attaches[$name] = array(
            'path'    => $path,
            'name'    => preg_replace('/./e', '"%".dechex(ord("$0"))', $name),
            'subtype' => $subtypes
        );
        
        $this->_build_attaches = false;
        return true;
    }
    
    /**
    * Добавляем адресата в список КОМУ
    */
    function AddTo($mail, $name = '') {
        $this->_a_to[$mail] = $name;
        $this->_build_to    = false;
    }
    
    /**
    * Удаляет адрес из списка КОМУ
    */
    function RemoveTo($mail) {
        unset($this->_a_to[$mail]);
        $this->_build_to    = false;
    }
    
    /**
    * Очищает список КОМУ
    */
    function ClearTo() {
        $this->_a_to        = array();
        $this->_build_to    = false;
    }
    
    /**
    * Установливает адрес, с которого придёт письмо (ОТ КОГО)
    */
    function SetFrom($mail, $name = '') {
        $this->_from            = array($mail, $name);
        $this->_build_headers   = false;
    }
    
    /**
    * Добавляем адресата в список КОПИЯ
    */
    function AddCC($mail, $name = '') {
        $this->_a_cc[$mail]     = $name;
        $this->_build_headers   = false;
    }
    
    /**
    * Удаляет адрес из списка КОПИЯ
    */
    function RemoveCC($mail) {
        unset($this->_a_cc[$mail]);
        $this->_build_headers = false;
    }
    
    /**
    * Очищает список КОПИЯ
    */
    function ClearCC() {
        $this->_a_cc            = array();
        $this->_build_headers   = false;
    }
    
    /**
    * Добавляем адресата в список СКРЫТАЯ КОПИЯ
    */
    function AddBCC($mail, $name = '') {
        $this->_a_bcc[$mail]    = $name;
        $this->_build_headers   = false;
    }
    
    /**
    * Удаляет адрес из списка СКРЫТАЯ КОПИЯ
    */
    function RemoveBCC($mail) {
        unset($this->_a_bcc[$mail]);
        $this->_build_headers = false;
    }
    
    /**
    * Очищает список СКРЫТАЯ КОПИЯ
    */
    function ClearBCC() {
        $this->_a_bcc           = array();
        $this->_build_headers   = false;
    }
    
    /**
    * Устанавливает адресата в заголовке REPLY-TO
    */
    function SetReplyTo($mail, $name = '') {
        $this->_reply_to        = array($mail, $name);
        $this->_build_headers   = false;
    }
    
    /**
    * Очищает заголовок REPLY-TO
    */
    function ClearReplyTo() {
        $this->_reply_to        = array();
        $this->_build_headers   = false;
    }
    
    /**
    * Устанавливает адресата в заголовке RETURN-PATH
    */
    function SetReturnPath($mail) {
        $this->AddHeader('Return-Path', $mail);
    }
    
    /**
    * Очищает заголовок RETURN-PATH
    */
    function ClearReturnPath() {
        RemoveHeader('Return-Path');
    }
    
    /**
    * Добавляет произвольный заголовок
    */
    function AddHeader($name, $value) {
        $this->_arr_headers[$name]  = trim($value);
        $this->_build_headers       = false;
    }
    
    /**
    * Удаляет указанный заголовок
    */
    function RemoveHeader($name) {
        unset($this->_arr_headers[$name]);
        $this->_build_headers  = false;
    }
    
    /**
     * Отсылает письмо
     *
     * @param boolean $ts_in_headers указание «подцепить» To и Subject к общему массиму $this->__headers
     */
    function Send($ts_in_headers = false) {
        if ( ! ($this->_build_body &&
            $this->_build_images &&
            $this->_build_attaches))    $this->_BuildBody();
        if ( ! $this->_build_to)        $this->_BuildTo();
        if ( ! $this->_build_subject)   $this->_BuildSubject();
        if ( ! $this->_build_headers)   $this->_BuildHeaders($ts_in_headers);
        
        // если заголовки не скомпонованы, письмо отослать невозможно
        if (empty($this->_build_headers)) return false;
        
        $st = $ts_in_headers
            ? mail('', '', $this->__message, $this->__headers)
            : mail($this->__to, $this->__subject, $this->__message, $this->__headers);
        
        // ведём статистику отосланных писем
        if ($st) ++$this->_sended;
        else ++$this->_nosended;
        
        return $st;
    }
    
    private function _BuildBody() {
        $rn = "\r\n";
        
        if ( ! $this->_build_body) {
            $this->__body = $this->_Base64Chunked($this->_body);
            $this->_build_body = true;
        }
        // если нет вложений, формируем простое письмо
        if ( ! ($this->_a_images || $this->_a_attaches)) {
            $this->__message = $this->__body;
            return;
        }
        
        $this->__message = 'This is a multi-part message in MIME format.' . $rn;
        $this->__message .= '--' . $this->_related . $rn
                          . 'Content-Type: text/'
                          . ($this->_type==='HTML' ? 'html' : 'plain')
                          . '; charset=utf-8' . $rn
                          . 'Content-Transfer-Encoding: base64' . $rn . $rn;
        $this->__message .= $this->__body . $rn;
        
        // если картинки или аттачи требует пересборки, выполним её
        
        if ( ! $this->_build_images) {
            foreach ($this->_a_images as $k => $v) {
                if ( ! isset($v['data'])) {
                    // получим содержимое файла
                    $t = file_get_contents($v['path']);
                    $this->_a_images[$k]['data'] =  $this->_Base64Chunked($t);
                }
            }
            $this->_build_images = true;
        }
        
        if ( ! $this->_build_attaches) {
            foreach ($this->_a_attaches as $k => $v) {
                // получим содержимое файла
                $t = file_get_contents($v['path']);
                $this->_a_attaches[$k] ['data']=  $this->_Base64Chunked($t);
            }
            $this->_build_attaches = true;
        }
        
        foreach ($this->_a_images as $cid => $vs) {
            $this->__message .= '--' . $this->_related . $rn
                . 'Content-Type: ' . $vs['subtype'] . ';'.$rn
                . ' name*=UTF-8\'\'' . $vs['name'] . $rn
                . 'Content-Transfer-Encoding: base64' . $rn
                . 'Content-Id: ' . $cid . $rn
                . $rn;
            $this->__message .= $vs['data'] . $rn;
        }
        
        foreach ($this->_a_attaches as $name => $vs) {
            $this->__message .= '--' . $this->_related . $rn
                . 'Content-Type: ' . $vs['subtype'] . ';'.$rn
                . ' name*=UTF-8\'\'' . $vs['name'] . $rn
                . 'Content-Transfer-Encoding: base64' . $rn
                . 'Content-Disposition: inline;' . $rn
                . ' filename*=UTF-8\'\'' . $vs['name'] . $rn
                . $rn;
            $this->__message .= $vs['data'] . $rn;
        }
        
        $this->__message .= '--' . $this->_related . '--' . $rn;
    }
    
    /**
    * Формирует тему письма
    */
    private function _BuildSubject() {
        // если тема не задана, будет использоваться пустая строка
        $this->_build_subject = true;
        if (empty($this->_subject)) $this->__subject = '';
        else {
            $this->__subject = $this->_MimeStr(implode(' ', $this->_MimeHeader($this->_subject)));
        }
    }
    
    /**
    * Формирует список адресатов
    */
    private function _BuildTo() {
        // обязательно должен быть задан адрес назначения
        if (empty($this->_a_to)) return; // сохраняем $this->_build_to = false;
        $txt = array();
        foreach ($this->_a_to as $mail => $name) {
            if (empty($name)) $txt[] = $mail;
            else $txt[] = implode(' ', $this->_MimeHeader($name)) . ' <' . $mail . '>';
        }
        $this->__to = $this->_MimeStr(implode(', ', $txt));
        $this->_build_to = true;
    }
    
    /**
    * Формирует массив заголовков
    *
    * @param boolean $ts_in_headers если true, то к заголовкам подцепляются также To и Subject
    */
    private function _BuildHeaders($ts_in_headers) {
        // cc, bcc, from, reply-to, return-path, headers
        
        $this->__headers = '';
        $rn = "\r\n";
        
        if ($ts_in_headers) {
            $this->__headers .= 'To: ' . $this->__to . $rn;
            $this->__headers .= 'Subject: ' . $this->__subject . $rn;
        }
        
        if ($this->_a_cc) {
            $txt = array();
            foreach ($this->_a_cc as $mail => $name) {
                if (empty($name)) $txt[] = $mail;
                else $txt[] = implode(' ', $this->_MimeHeader($name)) . ' <'
                            . $mail . '>';
            }
            $this->__headers .= $this->_MimeStr('CC: ' . implode(', ', $txt))
                              . $rn;
        }
        
        if ($this->_a_bcc) {
            $txt = array();
            foreach ($this->_a_bcc as $mail => $name) {
                if (empty($name)) $txt[] = $mail;
                else $txt[] = implode(' ', $this->_MimeHeader($name)) . ' <'
                            . $mail . '>';
            }
            $this->__headers .= $this->_MimeStr('BCC: ' . implode(', ', $txt))
                              . $rn;
        }
        
        if ($this->_from) {
            $txt = '';
            if (empty($this->_from[1])) $txt = $this->_from[0];
            else $txt = implode(' ', $this->_MimeHeader($this->_from[1]))
                        . ' <' . $this->_from[0] . '>';
            $this->__headers .= $this->_MimeStr('From: ' . $txt) . $rn;
        }
        
        if ($this->_reply_to) {
            $txt = '';
            if (empty($this->_reply_to[1])) $txt = $this->_reply_to[0];
            else $txt = implode(' ', $this->_MimeHeader($this->_reply_to[1]))
                        . ' <' . $this->_reply_to[0] . '>';
            $this->__headers .= $this->_MimeStr('Reply-To: ' . $txt) . $rn;
        }
        
        if ($this->_return_path) {
            $this->__headers .= 'Reply-To: ' . $this->_return_path . $rn;
        }
        
        // дополнительные заголовки, заданные администратором
        foreach ($this->_a_headers as $header => $value) {
            $this->__headers .= $header . ': ' . $value . $rn;
        }
        
        // @ используется на случай, если не установлена временная зона
        $this->__headers .= 'Date: ' . @date('r') . $rn; // дата по RFC 822
        
        $this->__headers .= 'Message-ID: <' . microtime(1) . '@'
                         . $_SERVER['HTTP_HOST'] . '>' . $rn;
        
        $this->__headers .= 'MIME-Version: 1.0'. $rn;
        
        // если есть вложения, добавляем значение разделителя границ письма
        if ($this->_a_images || $this->_a_attaches) {
            $this->__headers .= 'Content-Type: multipart/related;' . $rn // вместо mixed было related?!!
                             . ' boundary="' . $this->_related . '"' . $rn;
        } elseif ($this->_type == 'HTML') {
            $this->__headers .= 'Content-Type: text/html; charset=utf-8' . $rn
                             . 'Content-Transfer-Encoding: base64' . $rn;
        } else {
            $this->__headers .= 'Content-Type: text/plain; charset=utf-8' . $rn
                             . 'Content-Transfer-Encoding: base64' . $rn;
        }
        
        $this->_build_headers = true;
    }
    
    /**
    * Возвращает массив, компонентами которого являются части строки,
    * закодированной для MIME-шапки
    */
    private function _MimeHeader($s) {
        $txt = array();
        foreach (explode("\n", wordwrap($s, $this->CHUNK_LENGTH, "\n", 1)) as $v) {
            if ($v !== '') $txt[] = '=?UTF-8?B?' . base64_encode($v) . '?=';
        }
        return $txt;
    }
    
    /**
    * Возвращает строку, закодированную для MIME-шапки, со специфицированной
    * длинной строки
    */
    private function _MimeStr($txt) {
        return wordwrap($txt, $this->MAX_STRING_LENGTH, "\n ");
    }
    
    /**
    * Конвертирует переданную строку в base64, разбивает её на строки
    * специфицированной ширины и возвращает получившуюся строку.
    */
    private function _Base64Chunked($txt) {
        $rn = "\r\n";
        return wordwrap(base64_encode($txt), $this->MAX_STRING_LENGTH, $rn, 1);
    }
    
    // следующие 2 метода «перегружают» свойства объекта
    
    function __set($n, $v) {
        $t = '_'.$n;
        switch ($t) {
          case '_body':
            $this->$t = $v;
            $this->_build_body = false;
            break;
          
          case '_subject':
            $this->$t = strtr($v, "\r\n\t\0", '    ');
            $this->_build_subject = false;
            break;
        }
    }
    
    function __get($n) {
        $t = '_'.$n;
        switch ($t) {
          case '_body':
          case '_sended':
          case '_nosended':
          case '_subject':
            return $this->$t;
        }
    }
}
?>